# mch_cl_1.py - 2009 Jul 14
# Usage:
#    $ python mch_cl_1.py <CMD> <ARGS>
#    <CMD> = three-letter command TYPE (e.g., PNG)
#    <ARGS> = space-delimited list of arguments; depends on command
# Crude "stub"/"alpha" implementation of the MCS Common ICD from the client (MCS) side
# -- PNG
# -- RPT (but only one MIB entry at a time; not branches)
# -- SHT
# -- INI
# -- TMP
# -- DIF
# -- PWR
# The other end is the controlled subsystem (SHL or "server"); runs mch_shl_1.py
# This code uses the loopback IP for DEST_IP. To run it on a network, change it to the server's IP address.
# This is a "one shot" code -- it does it once and then quits.
# When run, it reads MIB_cl_1.txt, which represents the MIB in the following format:
#    index (space) label (space) value <CR> 
# The command provides information about what is happening as it goes.
# The command overwrites MIB_mincl_1.txt with an updated MIB, if appropriate.
# Some notes: 
# -- Note receive operation blocks, so execution will hang if server doesn't respond
# -- Intended to be compliant with MCS Common ICD v.1.0, except as noted above
# -- REFERENCE is currently fixed always to be 1391
# -- This is not production code.  

import socket
import time
import datetime
import math
import string
import struct   # for packing of binary to/from strings
import sys

DEST_IP = '127.0.0.1' # The IP address of the server
DEST_NAME = 'SHL'       # The 3-character MCS Common ICD designator for the server
PORTT = 1738             # The port address
PORTR = 1739             # The port address

# Below are things that shouldn't be changed
B = 8192                    # [bytes] Max message size
MIB_FILE = 'MIB_cl_1.txt' # Name of file containing MIB

#==============================
# Message Processing Subroutine
#==============================

def command(iref,cmd,data):
    # (in)  iref:    [int] message REFERENCE number to use 
    # (in)  cmd:     [str] three-letter command; e.g., "PNG"
    # (in)  data:    [str] DATA field of message
    # (out) err:     [int] Error code: sum of the following:
    #                   0: no errors
    #                   1: (not currently used)
    #                   2: wrong "DESTINATION"
    #                   4: R-RESPONSE = "R" (command reject indicator) received
    # (out) summary: [str] R-SUMMARY (i.e., sender's MIB 1.1)
    # (out) rdata:   [str] Contents of DATA field with R-* fields stripped away

    # ---------------------
    # Set up sockets
    # ---------------------

    # Set up the receive socket for UDP
    r = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    r.bind(('', PORTR))  # Accept connections from anywhere
    r.setblocking(1)    # Blocking on this socket 
    r.settimeout(3)     # Set timeout to 3 s

    # Set up transmit socket for UDP
    t = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    t.connect((DEST_IP, PORTT))

    # -------------------
    # Message Preparation
    # -------------------

    # determine current time
    dt = datetime.datetime.utcnow()
    year        = dt.year
    month       = dt.month
    day         = dt.day
    hour        = dt.hour
    minute      = dt.minute
    second      = dt.second
    millisecond = dt.microsecond / 1000

    # construct the MJD field
    # adapted from http://paste.lisp.org/display/73536
    # can check result using http://www.csgnetwork.com/julianmodifdateconv.html
    a = (14 - month) // 12
    y = year + 4800 - a
    m = month + (12 * a) - 3
    p = day + (((153 * m) + 2) // 5) + (365 * y)
    q = (y // 4) - (y // 100) + (y // 400) - 32045
    mjdi = int(math.floor( (p+q) - 2400000.5))
    mjd = string.rjust(str(mjdi),6)
    #print '#'+mjd+'#'

    # construct the MPM field
    mpmi = int(math.floor( (hour*3600 + minute*60 + second)*1000 + millisecond ) )
    mpm = string.rjust(str(mpmi),9) 
    #print '#'+mpm+'#'

    # Build the message
    # Note we are just using a single, non-updating REFERENCE number in this case
    message =           DEST_NAME
    message = message + 'MCS'
    message = message + cmd
    message = message + string.rjust(str(iref),9)
    message = message + string.rjust(str(len(data)),4)
    message = message + str(mjd)
    message = message + str(mpm) + ' '  # note trailing space
    message = message + data
    #print '#'+payload+'#' 

    # ---------------------
    # Sending and Receiving
    # ---------------------

    # Send it and wait for response
    t.send(message)            # send the message
    print 'sent> '+message+'|' # say what was sent
    #response = ''
    response = r.recv(B)       # this blocks until the server responds

    r.close() # close receive socket
    t.close() # close transmit socket

    # ---------------------
    # Analyzing response
    # ---------------------

    # parse message into payload and sent checksum
    message = response        # the message

    # say what was received
    print 'rcvd> '+message+'|'
    
    bDest     = False     # assume this until noted otherwise 
    bResponse = False     # assume this until noted otherwise
    summary   = 'UNKNOWN' # use this until we have a proper value
    rdata     = ''        # use this until we have a proper value 

    # check for correct recipient (DESTINATION)
    destination = message[:3]
    #print 'DESTINATION=|'+destination+'|'
    if destination=='MCS':
        bDest = True

        # check for command rejection (R-RESPONSE)
        rresponse = message[38:39]
        #print 'R-RESPONSE=|'+rresponse+'|'
        if rresponse=='A':
            bResponse = True

            # get R-SUMMARY and any remaining data
            summary = string.strip(message[39:46])
            rdata = string.strip(message[47:])
           
    # generate error code
    err = 0
    if not(bDest):
       err = err+2
    if not(bResponse):
       err = err+4

    # return
    return err, summary, rdata


#==================
# Main Program 
#==================

# Check for required command line argument <CMD> 
if len(sys.argv)<2: # no command specified -- crash out
    print 'Proper usage is "mch_minc_1.py <CMD> <ARGS>.  <CMD> missing.'
    print 'No action taken.'
    exit()
else:               # it's here -- use it
    cmd = string.strip(sys.argv[1])
    #print '|'+mib_label+'|'

iref = 1391  # hardwired for now

eErr = 0 # 0 = OK.  >0 represents an error code. 
         # 1 = checksum error in received message

#---------------
# Load the MIB 
#---------------

# Read the MIB
fp = open(MIB_FILE,'r')
file = fp.read()
fp.close()
#print file

# Initialize MIB variables
mib_index = []
mib_label = []
mib_value = []
line = file.splitlines() # split file into lines
nmib = len(line)         # number of MIB entries
for i in range(nmib):
    column = string.split(line[i],' ',2) # split into columns 
    mib_index.append(string.strip(column[0]))
    mib_label.append(column[1])
    mib_value.append(string.strip(column[2]))


#------------------
# Execute command 
#------------------

if cmd=='PNG':

    err, summary, rdata = command(iref,cmd,'')
    print 'Error Code (0 is good): '+str(err)
    mib_value[0] = summary # this works because summary=='UNKNOWN' if there is an error 

elif cmd=='RPT':

    # Check for required command line argument <CMD> 
    if len(sys.argv)<3: # no arguments specified -- crash out
        print 'Proper usage is "mch_minc_1.py RPT <MIB_LABEL>.  <MIB_LABEL> missing.'
        print 'No action taken.'
        exit()
    else:               # it's here -- use it
        arg = string.strip(sys.argv[2])
        # print '|'+arg+'|'

    # Is this really a MIB label?
    imib = -1
    for i in range(nmib):
        
        if arg==mib_label[i]:
            imib = i 
            
    if imib==-1:
        print 'MIB_LABEL |'+arg+'| not found.'
        print 'No action taken.'
        exit()
       
    err, summary, rdata = command(iref,cmd,arg)
    print 'Error Code (0 is good): '+str(err)
    mib_value[0] = summary # this works because summary=='UNKNOWN' if there is an error 
       
    if err==0:
        mib_value[imib] = rdata 

elif cmd=='SHT':

    # Check for additional command line arguments
    arg='' 
    for i in range(2,len(sys.argv)):
       if sys.argv[i]=='SCRAM':
           arg = arg+'SCRAM '
       elif sys.argv[i]=='RESTART':
           arg = arg+'RESTART '
       else:
           print 'I do not understand |'+sys.argv[i]+'|'
           print 'No action taken.'
           exit()
    arg = string.strip(arg)
    # print '|'+arg+'|'
       
    err, summary, rdata = command(iref,cmd,arg)
    print 'Error Code (0 is good): '+str(err)
    mib_value[0] = summary # this works because summary=='UNKNOWN' if there is an error 
   
elif cmd=='INI':
    
    # Check for additional command line arguments
    arg=''
    if (len(sys.argv)<5):
        print 'Proper usage is "mod_client.py INI <Set-Point> <Diff Set Point> <Racks Installed>" | Enter appropriate number of arguments'
        print 'No action taken'
        exit()
    else:
        arg1 = string.strip(sys.argv[2])
        if (len(arg1)>5):
             print 'The Set-Point should not be larger than 5 bytes'
             print 'No action taken' 
             exit()       
        if (len(arg1)<5):
            arg1 = string.rjust(arg1,5,'0')  
        arg2 = string.strip(sys.argv[3])
        if (len(arg2)>3):
             print 'The Differential Set-Point should not be larger than 3 bytes| Should be between 0.5 F and 5 F'
             print 'No action taken' 
             exit()       
        if (len(arg2)<3):
             arg2 = string.rjust(arg2,3,'0')  
        arg3 = string.strip(sys.argv[4])

        if (len(arg3)!=6):
             print ' "No: of racks installed" argument format incorrect. Enter 6 bytes'
             print 'No action taken'
             exit()
                
        arg = arg + arg1 +'&'+ arg2 + '&'+arg3
  

    err, summary, rdata = command(iref,cmd,arg)
    print 'Error Code (0 is good): '+str(err)
    mib_value[0] = summary

elif cmd == 'TMP':
     arg = ''
     if (len(sys.argv)<3):
         print 'Proper usage is "mod_client.py TMP <Set-Point>" | Enter appropriate number of arguments'
         print 'No action taken'
         exit()
     else:
         arg1 = string.strip(sys.argv[2])
         if (len(arg1)>5):
             print 'The Set-Point should not be larger than 5 bytes'
             print 'No action taken' 
             exit()       
         if (len(arg1)<5):
             arg1 = string.rjust(arg1,5,'0')  
         arg = arg + arg1
         # print '|'+arg+'|'

         err, summary, rdata = command(iref,cmd,arg)
         print 'Error Code (0 is good): '+str(err)
         mib_value[0] = summary

elif cmd == 'DIF':
     arg = ''
     if (len(sys.argv)<3):
         print 'Proper usage is "mod_client.py DIF <Differential Set-Point>" | Enter appropriate number of arguments'
         print 'No action taken'
         exit()
     else:
         arg1 = string.strip(sys.argv[2])
         if (len(arg1)>3):
             print 'The Differential Set-Point should not be larger than 3 bytes| Should be between 0.5 F and 5 F'
             print 'No action taken' 
             exit()       
         if (len(arg1)<3):
             arg1 = string.rjust(arg1,3,'0')  
         
         arg = arg + arg1
             
         # print '|'+arg+'|'

         err, summary, rdata = command(iref,cmd,arg)
         print 'Error Code (0 is good): '+str(err)
         mib_value[0] = summary

elif cmd == 'PWR':
     arg =''
     if (len(sys.argv)<5):
         print 'Proper usage is "mod_client.py PWR <Rack Number> <Port Number> <Control(On/OFF)>" | Enter appropriate    number of arguments'
         print 'No action taken'
         exit()
     rack = string.strip(sys.argv[2])
     port = string.strip(sys.argv[3])
     control = string.strip(sys.argv[4])
     if (len(rack) == 1):
         if (int(rack)>=1 and int(rack)<=6):
             arg = arg + rack
         else:
             print 'Invalid rack entry | Enter values between 1 and 6'
             print 'No action taken'
             exit()
     else: 
          print 'The rack entry should not be larger than a byte'
          print 'No action taken'
          exit()
     if (len(port)>2):
         print 'The port entry should not be larger than 2 bytes | Should be between 1 and 50'
         print 'No action taken'
         exit()
     else: 
         check = float(port)
         if (check == int(port)):
            if (int(port)>=1 and int(port)<=50):
               if (len(port)==1):
                   port = string.rjust(port,2,'0')
               arg = arg + port              
            else:
               print 'Invalid port entry | Should be between 1 and 50'
               print 'No action taken'
               exit()
         else:
            print 'Enter only valid integer values for the ports'
            print 'No action taken'
            exit()
         
     if (len(control)>3):
         print 'Invalid Control argument'
         print 'No action taken'
         exit()
     if (control == 'ON' or control == 'OFF'):
         if (control == 'ON'):
             control = 'ON '
         arg = arg + control
         # print '#'+arg+'#'
     else:
          print 'Invalid control argument'
          print 'No action taken'
          exit()
     # print '|'+arg+'|'

     err, summary, rdata = command(iref,cmd,arg)
     print 'Error Code (0 is good): '+str(err)
     mib_value[0] = summary                  
                
           
            
else:
    print 'I do not recognize the command <'+cmd+'>.'
    print 'No action taken.'
    exit()

#---------------
# Update MIB 
#---------------

# Overwrite the MIB file
if eErr==0:
    fp= open(MIB_FILE,'w')
    for i in range(nmib):
        fp.write(mib_index[i]+' '+mib_label[i]+' '+mib_value[i]+'\n')
    fp.close()










